# -*- coding: utf-8 -*-

# Public API
from .client import Client
from .oauth_client import OAuthClient
from .query_builder import QueryBuilder
from .resource import Resource
from .params_builder import ParamsBuilder

__author__ = "Robert Wikman <rbw@vault13.org>"
__version__ = "0.7.3"

# Set default logging handler to avoid "No handler found" warnings.
import logging
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())



# -*- coding: utf-8 -*-

import inspect
import six

from .exceptions import (QueryEmpty,
                         QueryExpressionError,
                         QueryMissingField,
                         QueryMultipleExpressions,
                         QueryTypeError)


class QueryBuilder(object):
    """Query builder - for constructing advanced ServiceNow queries"""
    def __init__(self):
        self._query = []
        self.current_field = None
        self.c_oper = None
        self.l_oper = None

    def field(self, field):
        """Sets the field to operate on

        :param field: field (str) to operate on
        """

        self.current_field = field
        return self

    def order_descending(self):
        """Sets ordering of field descending"""

        self._query.append('ORDERBYDESC{0}'.format(self.current_field))
        self.c_oper = inspect.currentframe().f_back.f_code.co_name
        return self

    def order_ascending(self):
        """Sets ordering of field ascending"""

        self._query.append('ORDERBY{0}'.format(self.current_field))
        self.c_oper = inspect.currentframe().f_back.f_code.co_name
        return self

    def starts_with(self, starts_with):
        """Adds new `STARTSWITH` condition

        :param starts_with: Match field starting with the provided value
        """

        return self._add_condition('STARTSWITH', starts_with, types=[str])

    def ends_with(self, ends_with):
        """Adds new `ENDSWITH` condition

        :param ends_with: Match field ending with the provided value
        """

        return self._add_condition('ENDSWITH', ends_with, types=[str])

    def contains(self, contains):
        """Adds new `LIKE` condition

        :param contains: Match field containing the provided value
        """

        return self._add_condition('LIKE', contains, types=[str])

    def not_contains(self, not_contains):
        """Adds new `NOTLIKE` condition

        :param not_contains: Match field not containing the provided value
        """

        return self._add_condition('NOTLIKE', not_contains, types=[str])

    def is_empty(self):
        """Adds new `ISEMPTY` condition"""

        return self._add_condition('ISEMPTY', '', types=[str, int])

    def equals(self, data):
        """Adds new `IN` or `=` condition depending on if a list or string was provided

        :param data: string or list of values
        :raise:
            - QueryTypeError: if `data` is of an unexpected type
        """

        if isinstance(data, six.string_types):
            return self._add_condition('=', data, types=[int, str])
        elif isinstance(data, list):
            return self._add_condition('IN', ",".join(map(str, data)), types=[str])

        raise QueryTypeError('Expected value of type `str` or `list`, not %s' % type(data))

    def not_equals(self, data):
        """Adds new `NOT IN` or `!=` condition depending on if a list or string was provided

        :param data: string or list of values
        :raise:
            - QueryTypeError: if `data` is of an unexpected type
        """

        if isinstance(data, six.string_types):
            return self._add_condition('!=', data, types=[int, str])
        elif isinstance(data, list):
            return self._add_condition('NOT IN', ",".join(data), types=[str])

        raise QueryTypeError('Expected value of type `str` or `list`, not %s' % type(data))

    def greater_than(self, greater_than):
        """Adds new `>` condition

        :param greater_than: str or datetime compatible object
        :raise:
            - QueryTypeError: if `greater_than` is of an unexpected type
        """

        if hasattr(greater_than, 'strftime'):
            greater_than = greater_than.strftime('%Y-%m-%d %H:%M:%S')
        elif isinstance(greater_than, six.string_types):
            raise QueryTypeError('Expected value of type `int` or instance of `datetime`, not %s' % type(greater_than))

        return self._add_condition('>', greater_than, types=[int, str])

    def less_than(self, less_than):
        """Adds new `<` condition

        :param less_than: str or datetime compatible object
        :raise:
            - QueryTypeError: if `less_than` is of an unexpected type
        """

        if hasattr(less_than, 'strftime'):
            less_than = less_than.strftime('%Y-%m-%d %H:%M:%S')
        elif isinstance(less_than, six.string_types):
            raise QueryTypeError('Expected value of type `int` or instance of `datetime`, not %s' % type(less_than))

        return self._add_condition('<', less_than, types=[int, str])

    def between(self, start, end):
        """Adds new `BETWEEN` condition

        :param start: int or datetime  compatible object
        :param end: int or datetime compatible object
        :raise:
            - QueryTypeError: if start or end arguments is of an invalid type
        """

        if hasattr(start, 'strftime') and hasattr(end, 'strftime'):
            dt_between = (
              'javascript:gs.dateGenerate("%(start)s")'
              "@"
              'javascript:gs.dateGenerate("%(end)s")'
            ) % {
              'start': start.strftime('%Y-%m-%d %H:%M:%S'),
              'end': end.strftime('%Y-%m-%d %H:%M:%S')
            }
        elif isinstance(start, int) and isinstance(end, int):
            dt_between = '%d@%d' % (start, end)
        else:
            raise QueryTypeError("Expected `start` and `end` of type `int` "
                                 "or instance of `datetime`, not %s and %s" % (type(start), type(end)))

        return self._add_condition('BETWEEN', dt_between, types=[str])

    def AND(self):
        """Adds an and-operator"""
        return self._add_logical_operator('^')

    def OR(self):
        """Adds an or-operator"""
        return self._add_logical_operator('^OR')

    def NQ(self):
        """Adds a NQ-operator (new query)"""
        return self._add_logical_operator('^NQ')

    def _add_condition(self, operator, operand, types):
        """Appends condition to self._query after performing validation

        :param operator: operator (str)
        :param operand: operand
        :param types: allowed types
        :raise:
            - QueryMissingField: if a field hasn't been set
            - QueryMultipleExpressions: if a condition already has been set
            - QueryTypeError: if the value is of an unexpected type
        """

        if not self.current_field:
            raise QueryMissingField("Conditions requires a field()")

        elif not type(operand) in types:
            caller = inspect.currentframe().f_back.f_code.co_name
            raise QueryTypeError("Invalid type passed to %s() , expected: %s" % (caller, types))

        elif self.c_oper:
            raise QueryMultipleExpressions("Expected logical operator after expression")

        self.c_oper = inspect.currentframe().f_back.f_code.co_name

        self._query.append("%(current_field)s%(operator)s%(operand)s" % {
                               'current_field': self.current_field,
                               'operator': operator,
                               'operand': operand
        })

        return self

    def _add_logical_operator(self, operator):
        """Adds a logical operator in query

        :param operator: logical operator (str)
        :raise:
            - QueryExpressionError: if a expression hasn't been set
        """

        if not self.c_oper:
            raise QueryExpressionError("Logical operators must be preceded by an expression")

        self.current_field = None
        self.c_oper = None

        self.l_oper = inspect.currentframe().f_back.f_code.co_name
        self._query.append(operator)
        return self

    def __str__(self):
        """String representation of the query object

        :raise:
            - QueryEmpty: if there's no conditions defined
            - QueryMissingField: if field() hasn't been set
            - QueryExpressionError: if a expression hasn't been set

        :return:
            - String-type query
        """

        if len(self._query) == 0:
            raise QueryEmpty("At least one condition is required")
        elif self.current_field is None:
            raise QueryMissingField("Logical operator expects a field()")
        elif self.c_oper is None:
            raise QueryExpressionError("field() expects an expression")

        return str().join(self._query)



# -*- coding: utf-8 -*-

import os
from pysnow.exceptions import InvalidUsage


class Attachment(object):
    """Attachment management

    :param resource: Table API resource to manage attachments for
    :param table_name: Name of the table to use in the attachment API
    """

    def __init__(self, resource, table_name):
        self.resource = resource
        self.table_name = table_name

    def get(self, sys_id=None, limit=100):
        """Returns a list of attachments

        :param sys_id: record sys_id to list attachments for
        :param limit: override the default limit of 100
        :return: list of attachments
        """

        if sys_id:
            return self.resource.get(query={'table_sys_id': sys_id, 'table_name': self.table_name}).all()

        return self.resource.get(query={'table_name': self.table_name}, limit=limit).all()

    def upload(self, sys_id, file_path, name=None, multipart=False):
        """Attaches a new file to the provided record

        :param sys_id: the sys_id of the record to attach the file to
        :param file_path: local absolute path of the file to upload
        :param name: custom name for the uploaded file (instead of basename)
        :param multipart: whether or not to use multipart
        :return: the inserted record
        """

        if not isinstance(multipart, bool):
            raise InvalidUsage('Multipart must be of type bool')

        resource = self.resource

        if name is None:
            name = os.path.basename(file_path)

        resource.parameters.add_custom({
            'table_name': self.table_name,
            'table_sys_id': sys_id,
            'file_name': name
        })

        data = open(file_path, 'rb').read()
        headers = {}

        if multipart:
            headers["Content-Type"] = "multipart/form-data"
            path_append = '/upload'
        else:
            headers["Content-Type"] = "text/plain"
            path_append = '/file'

        return resource.request(method='POST', data=data, headers=headers, path_append=path_append)

    def delete(self, sys_id):
        """Deletes the provided attachment record

        :param sys_id: attachment sys_id
        :return: delete result
        """

        return self.resource.delete(query={'sys_id': sys_id})



# -*- coding: utf-8 -*-

import ijson

from ijson.common import ObjectBuilder
from itertools import chain
from .exceptions import (ResponseError,
                         NoResults,
                         InvalidUsage,
                         MultipleResults,
                         MissingResult)


class Response(object):
    """Takes a :class:`requests.Response` object and performs deserialization and validation.

    :param response: :class:`requests.Response` object
    :param resource: parent :class:`resource.Resource` object
    :param chunk_size: Read and return up to this size (in bytes) in the stream parser
    """

    def __init__(self, response, resource, chunk_size=2048, stream=False):
        self._response = response
        self._chunk_size = chunk_size
        self._count = 0
        self._resource = resource
        self._stream = stream

    @property
    def count(self):
        return self._count

    @count.setter
    def count(self, count):
        if not isinstance(count, int) or isinstance(count, bool):
            raise TypeError("Count must be an integer")

        self._count = count

    def __getitem__(self, key):
        return self.one().get(key)

    def __repr__(self):
        return '<%s [%d - %s]>' % (self.__class__.__name__, self._response.status_code, self._response.request.method)

    def _parse_response(self):
        """Looks for `result.item` (array), `result` (object) and `error` (object) keys and parses
        the raw response content (stream of bytes)

        :raise:
            - ResponseError: If there's an error in the response
            - MissingResult: If no result nor error was found
        """

        response = self._get_response()

        has_result_single = False
        has_result_many = False
        has_error = False

        builder = ObjectBuilder()

        for prefix, event, value in ijson.parse(response.raw, buf_size=self._chunk_size):
            if (prefix, event) == ('error', 'start_map'):
                # Matched ServiceNow `error` object at the root
                has_error = True
            elif prefix == 'result' and event in ['start_map', 'start_array']:
                # Matched ServiceNow `result`
                if event == 'start_map':  # Matched object
                    has_result_single = True
                elif event == 'start_array':  # Matched array
                    has_result_many = True

            if has_result_many:
                # Build the result
                if (prefix, event) == ('result.item', 'end_map'):
                    # Reached end of object. Set count and yield
                    builder.event(event, value)
                    self.count += 1
                    yield getattr(builder, 'value')
                elif prefix.startswith('result.item'):
                    # Build the result object
                    builder.event(event, value)
            elif has_result_single:
                if (prefix, event) == ('result', 'end_map'):
                    # Reached end of the result object. Set count and yield.
                    builder.event(event, value)
                    self.count += 1
                    yield getattr(builder, 'value')
                elif prefix.startswith('result'):
                    # Build the error object
                    builder.event(event, value)
            elif has_error:
                if (prefix, event) == ('error', 'end_map'):
                    # Reached end of the error object - raise ResponseError exception
                    raise ResponseError(getattr(builder, 'value'))
                elif prefix.startswith('error'):
                    # Build the error object
                    builder.event(event, value)

        if (has_result_single or has_result_many) and self.count == 0:  # Results empty
            return

        if not (has_result_single or has_result_many or has_error):  # None of the expected keys were found
            raise MissingResult('The expected `result` key was missing in the response. Cannot continue')

    def _get_response(self):
        response = self._response

        # Raise an HTTPError if we hit a non-200 status code
        response.raise_for_status()

        return response

    def _get_streamed_response(self):
        """Parses byte stream (memory efficient)

        :return: Parsed JSON
        """

        yield self._parse_response()

    def _get_buffered_response(self):
        """Returns a buffered response

        :return: Buffered response
        """

        response = self._get_response()

        if response.request.method == 'DELETE' and response.status_code == 204:
            return [{'status': 'record deleted'}], 1

        result = self._response.json().get('result', None)

        if result is None:
            raise MissingResult('The expected `result` key was missing in the response. Cannot continue')

        length = 0

        if isinstance(result, list):
            length = len(result)
        elif isinstance(result, dict):
            result = [result]
            length = 1

        return result, length

    def all(self):
        """Returns a chained generator response containing all matching records

        :return:
            - Iterable response
        """

        if self._stream:
            return chain.from_iterable(self._get_streamed_response())

        return self._get_buffered_response()[0]

    def first(self):
        """Return the first record or raise an exception if the result doesn't contain any data

        :return:
            - Dictionary containing the first item in the response content

        :raise:
            - NoResults: If no results were found
        """

        if not self._stream:
            raise InvalidUsage('first() is only available when stream=True')

        try:
            content = next(self.all())
        except StopIteration:
            raise NoResults("No records found")

        return content

    def first_or_none(self):
        """Return the first record or None

        :return:
            - Dictionary containing the first item or None
        """

        try:
            return self.first()
        except NoResults:
            return None

    def one(self):
        """Return exactly one record or raise an exception.

        :return:
            - Dictionary containing the only item in the response content

        :raise:
            - MultipleResults: If more than one records are present in the content
            - NoResults: If the result is empty
        """

        result, count = self._get_buffered_response()

        if count == 0:
            raise NoResults("No records found")
        elif count > 1:
            raise MultipleResults("Expected single-record result, got multiple")

        return result[0]

    def one_or_none(self):
        """Return at most one record or raise an exception.

        :return:
            - Dictionary containing the matching record or None

        :raise:
            - MultipleResults: If more than one records are present in the content
        """

        try:
            return self.one()
        except NoResults:
            return None

    def update(self, payload):
        """Convenience method for updating a fetched record

        :param payload: update payload
        :return: update response object
        """

        return self._resource.update({'sys_id': self['sys_id']}, payload)

    def delete(self):
        """Convenience method for deleting a fetched record

        :return: delete response object
        """

        return self._resource.delete({'sys_id': self['sys_id']})

    def upload(self, *args, **kwargs):
        """Convenience method for attaching files to a fetched record

        :param args: args to pass along to `Attachment.upload`
        :param kwargs: kwargs to pass along to `Attachment.upload`
        :return: upload response object
        """

        return self._resource.attachments.upload(self['sys_id'], *args, **kwargs)



# -*- coding: utf-8 -*-

import logging
import inspect
import warnings

import requests
import pysnow

from requests.auth import HTTPBasicAuth
from .legacy_request import LegacyRequest
from .exceptions import InvalidUsage
from .resource import Resource
from .url_builder import URLBuilder
from .params_builder import ParamsBuilder

warnings.simplefilter("always", DeprecationWarning)

logger = logging.getLogger('pysnow')


class Client(object):
    """User-created Client object.

    :param instance: Instance name, used to construct host
    :param host: Host can be passed as an alternative to instance
    :param user: User name
    :param password: Password
    :param raise_on_empty: Whether or not to raise an exception on 404 (no matching records), defaults to True
    :param request_params: Request params to send with requests globally (deprecated)
    :param use_ssl: Enable or disable the use of SSL, defaults to True
    :param session: Optional :class:`requests.Session` object to use instead of passing user/pass to :class:`Client`
    :raises:
        - InvalidUsage: On argument validation error
    """

    def __init__(self,
                 instance=None,
                 host=None,
                 user=None,
                 password=None,
                 raise_on_empty=None,
                 request_params=None,
                 use_ssl=True,
                 session=None):

        if (host and instance) is not None:
            raise InvalidUsage("Arguments 'instance' and 'host' are mutually exclusive, you cannot use both.")

        if type(use_ssl) is not bool:
            raise InvalidUsage("Argument 'use_ssl' must be of type bool")

        if raise_on_empty is None:
            self.raise_on_empty = True
        elif type(raise_on_empty) is bool:
            warnings.warn("The use of the `raise_on_empty` argument is deprecated and will be removed in a "
                          "future release.", DeprecationWarning)

            self.raise_on_empty = raise_on_empty
        else:
            raise InvalidUsage("Argument 'raise_on_empty' must be of type bool")

        if not (host or instance):
            raise InvalidUsage("You must supply either 'instance' or 'host'")

        if not isinstance(self, pysnow.OAuthClient):
            if not (user and password) and not session:
                raise InvalidUsage("You must supply either username and password or a session object")
            elif (user and session) is not None:
                raise InvalidUsage("Provide either username and password or a session, not both.")

        self.parameters = ParamsBuilder()

        if request_params is not None:
            warnings.warn("The use of the `request_params` argument is deprecated and will be removed in a "
                          "future release. Please use Client.parameters instead.", DeprecationWarning)

            self.parameters.add_custom(request_params)

        self.request_params = request_params or {}
        self.instance = instance
        self.host = host
        self._user = user
        self._password = password
        self.use_ssl = use_ssl
        self.base_url = URLBuilder.get_base_url(use_ssl, instance, host)

        if not isinstance(self, pysnow.OAuthClient):
            self.session = self._get_session(session)
        else:
            self.session = None

    def _get_session(self, session):
        """Creates a new session with basic auth, unless one was provided, and sets headers.

        :param session: (optional) Session to re-use
        :return:
            - :class:`requests.Session` object
        """

        if not session:
            logger.debug('(SESSION_CREATE) User: %s' % self._user)
            s = requests.Session()
            s.auth = HTTPBasicAuth(self._user, self._password)
        else:
            logger.debug('(SESSION_CREATE) Object: %s' % session)
            s = session

        s.headers.update(
            {
                'content-type': 'application/json',
                'accept': 'application/json',
                'User-Agent': 'pysnow/%s' % pysnow.__version__
            }
        )

        return s

    def _legacy_request(self, method, table, **kwargs):
        """Returns a :class:`LegacyRequest` object, compatible with Client.query and Client.insert

        :param method: HTTP method
        :param table: Table to operate on
        :return:
            - :class:`LegacyRequest` object
        """

        warnings.warn("`%s` is deprecated and will be removed in a future release. "
                      "Please use `resource()` instead." % inspect.stack()[1][3], DeprecationWarning)

        return LegacyRequest(method,
                             table,
                             request_params=self.request_params,
                             raise_on_empty=self.raise_on_empty,
                             session=self.session,
                             instance=self.instance,
                             base_url=self.base_url,
                             **kwargs)

    def resource(self, api_path=None, base_path='/api/now', chunk_size=None):
        """Creates a new :class:`Resource` object after validating paths

        :param api_path: Path to the API to operate on
        :param base_path: (optional) Base path override
        :param chunk_size: Response stream parser chunk size (in bytes)
        :return:
            - :class:`Resource` object
        :raises:
            - InvalidUsage: If a path fails validation
        """

        for path in [api_path, base_path]:
            URLBuilder.validate_path(path)

        return Resource(api_path=api_path,
                        base_path=base_path,
                        parameters=self.parameters,
                        chunk_size=chunk_size or 8192,
                        session=self.session,
                        base_url=self.base_url)

    def query(self, table, **kwargs):
        """Query (GET) request wrapper.

        :param table: table to perform query on
        :param kwargs: Keyword arguments passed along to `Request`
        :return:
            - List of dictionaries containing the matching records
        """

        return self._legacy_request('GET', table, **kwargs)

    def insert(self, table, payload, **kwargs):
        """Insert (POST) request wrapper

        :param table: table to insert on
        :param payload: update payload (dict)
        :param kwargs: Keyword arguments passed along to `Request`
        :return:
            - Dictionary containing the created record
        """

        r = self._legacy_request('POST', table, **kwargs)
        return r.insert(payload)

